; --------------------------------------------------------------------------- ;
;                       Treeview Drag & Drop Application
; --------------------------------------------------------------------------- ;
;
; Author: dorf (John Wildenauer) www.phpmvc.net
; Date  : 14-Sept-2oo4
;
; Desc  : This project creates a Treeview control with drag and drop operation,
;         based on Iczelions Treeview tutorial (#19) [win32asm.cjb.net].
;         It also demonstrates how to change a Treeview item image on-the-fly
;         when the user left-clicks an item. This could be used to highlite
;         a drop candidate for the drag-and-drop operation.
;
;         In this demo, the "First Node" item is the only valid drop target. 
;         When the user clicks on any other node, the "First Node" image is
;         changed to indicate that the "First Node" is a drop target. If the
;         user drops an item on the "First Node" a confirmation is displayed
;         and the "First Node" is selected. In all other cases the drop is
;         ignored and the original (clicked) item is reselected.
;
; Issues: The "WM_LBUTTONUP" mouse-up in a Treeview item does not work as expected
;         We can send a custom "WM_MOUSEMOVE" message to the Treeview procedure
;         as a workaround. Not ideal but it works.
; --------------------------------------------------------------------------- ;

.586
.model flat,stdcall
option casemap:none


; Note: 
; The default MASM install path is usually \Masm32\...
; Change these paths to suit your system
include		windows.inc 
include		user32.inc 
include		kernel32.inc
;include		gdi32.inc
include		comctl32.inc 

includelib	user32.lib 
includelib	kernel32.lib
includelib	comctl32.lib 
;includelib	gdi32.lib

; MASM32 Library
include masm32.inc
includelib masm32.lib

; Our custom library
include		.\Lib\String.Inc

; Our application definitions etc
include .\AdvTreeview.Inc

InsertSon	proto :dword,:dword,:dword
IfIncest	proto :dword,:dword,:dword
; --------------------------------------------------------------------------- ;

.code

start:

	invoke GetModuleHandle, NULL
	mov    hInstance,eax

	invoke GetCommandLine
	mov    CommandLine,eax

	; Prepare common control structure
	mov		icex.dwSize,sizeof INITCOMMONCONTROLSEX
	mov		icex.dwICC,ICC_DATE_CLASSES or ICC_USEREX_CLASSES or ICC_INTERNET_CLASSES\
				 or ICC_ANIMATE_CLASS or ICC_HOTKEY_CLASS or ICC_PAGESCROLLER_CLASS\
				 or ICC_COOL_CLASSES
	invoke InitCommonControlsEx,addr icex

	invoke WinMain, hInstance,NULL,CommandLine, SW_SHOWDEFAULT
	invoke ExitProcess,eax


WinMain proc hInst:HINSTANCE,hPrevInst:HINSTANCE,CmdLine:LPSTR,CmdShow:DWORD
	LOCAL wc:WNDCLASSEX
	LOCAL msg:MSG
	LOCAL hwnd:HWND
	
	mov   wc.cbSize,SIZEOF WNDCLASSEX
	mov   wc.style, CS_HREDRAW or CS_VREDRAW
	mov   wc.lpfnWndProc, OFFSET WndProc
	mov   wc.cbClsExtra,NULL
	mov   wc.cbWndExtra,NULL
	push  hInstance
	pop   wc.hInstance
	mov   wc.hbrBackground,COLOR_BTNFACE+1
	mov   wc.lpszMenuName,NULL
	mov   wc.lpszClassName,OFFSET szClassName
	
	invoke LoadIcon, hInst, IDI_APP_ICON	; <<<<<<<<<<<<<<<
	mov   wc.hIcon,eax
	mov   wc.hIconSm,eax
	
	invoke LoadCursor,NULL,IDC_ARROW
	mov   wc.hCursor,eax
	
	invoke RegisterClassEx, addr wc
	INVOKE CreateWindowEx,NULL,ADDR szClassName,ADDR szAppName,\
           WS_OVERLAPPEDWINDOW,\
           CW_USEDEFAULT,CW_USEDEFAULT, 255, 350,\
           200,NULL,hInst,NULL
	mov   hwnd,eax


	; Build the Treeview control window
	invoke CreateWindowEx, 0, addr szTreeViewClass, NULL, 
				WS_CHILD or WS_VISIBLE or WS_CLIPCHILDREN or WS_CLIPSIBLINGS or \
				TVS_HASLINES or TVS_HASBUTTONS or TVS_SHOWSELALWAYS or TVS_TRACKSELECT,\
				10, 10, 225, 300,\
				hwnd, 0, hInstance, 0	
	mov		hPrjTrv, eax
	
	; Set the Treeview procedure
	invoke SetWindowLong, hPrjTrv, GWL_WNDPROC, offset TreeViewProc
	; And save the old Treeview procedure
	mov		OldTreeViewProc, eax
	

	; Setup the image list. (16*16 pix bitmaps, 7 off, zero extra bitmaps)
	invoke ImageList_Create, 16 ,16 , ILC_COLOR8 or ILC_MASK, 17, 0
	mov		ebx, eax
	invoke LoadBitmap, hInstance, IDB_PROTRV			; "Res/tv-docs16.bmp"
	invoke ImageList_AddMasked, ebx, eax, 0FFFFFFh	; Generate a mask from the specified bitmap
	mov	hPrjIml, ebx										; Save the imagelist handle
	invoke SendMessage, hPrjTrv, TVM_SETIMAGELIST, 0, hPrjIml
	

	; Add some Treeview nodes, with bitmap images
	invoke AddTreeViewNode, hPrjTrv, hTVItem00, hTVItem00, CStr("Root Node"), 0, 0, 0
	mov	hTVItem00, eax
	invoke AddTreeViewNode, hPrjTrv, hTVItem00, hTVItem00, CStr("First Node"), 1, 1, 0
	mov	hTVItem01, eax
	invoke AddTreeViewNode, hPrjTrv, hTVItem01, hTVItem00, CStr("Second Node"), 2, 2, 0
	mov	hTVItem02, eax
	invoke AddTreeViewNode, hPrjTrv, hTVItem00, hTVItem00, CStr("Third Node"), 3, 3, 0
	mov	hTVItem03, eax
	invoke AddTreeViewNode, hPrjTrv, hTVItem03, hTVItem00, CStr("Fourth Node"), 4, 4, 0
	mov	hTVItem04, eax

	; Expand the tree
	invoke SendMessage, hPrjTrv, TVM_EXPAND, TVE_EXPAND, hTVItem00
	invoke SendMessage, hPrjTrv, TVM_EXPAND, TVE_EXPAND, hTVItem01
	invoke SendMessage, hPrjTrv, TVM_EXPAND, TVE_EXPAND, hTVItem03

	; And select the root node
	invoke SendMessage, hPrjTrv, TVM_SELECTITEM, TVGN_CARET, hTVItem00


	invoke ShowWindow, hwnd,SW_SHOWNORMAL
	invoke UpdateWindow, hwnd
	
	.WHILE TRUE
		invoke GetMessage, ADDR msg,NULL,0,0
		.BREAK .IF (!eax)
		invoke TranslateMessage, ADDR msg
		invoke DispatchMessage, ADDR msg
	.ENDW
	
	mov     eax,msg.wParam
	ret
WinMain endp

WndProc proc hWnd:HWND, uMsg:UINT, wParam:WPARAM, lParam:LPARAM
	
	.IF uMsg==WM_CREATE
		;
	;.ELSEIF uMsg==WM_MOUSEMOVE
		;
	;.ELSEIF uMsg==WM_LBUTTONUP
		;
	.ELSEIF uMsg==WM_NOTIFY
		; Notify the Treeview window that an event has occurred in the control.
		invoke SendMessage, hPrjTrv, WM_NOTIFY, wParam, lParam
	.ELSEIF uMsg==WM_DESTROY 
		invoke PostQuitMessage,NULL
	.ELSE
		invoke DefWindowProc,hWnd,uMsg,wParam,lParam		
		ret
	.ENDIF
	
	xor eax,eax
	ret
WndProc endp


; TreeViewProc
;	Desc   : The Treeview Windows procedure
;	Class  : szTreeViewClass
; 	Handle : hPrjTrv
;	@param	hWin	The Treeview control window handle.
;
TreeViewProc proc hWin:HWND, uMsg:UINT, wParam:WPARAM, lParam:LPARAM
	LOCAL tvhit:TV_HITTESTINFO	; This structure is used with the TVM_HITTEST message
	LOCAL	tvi:TV_ITEMEX			; Specifies or receives attributes of a tree-view item.
	LOCAL	buffItemSrc[64]:BYTE	; The source Treeview item
	LOCAL	buffItemTgt[64]:BYTE	; The target Treeview item
	LOCAL	buffItemMsg[64]:BYTE	; A message buffer
	LOCAL hTVItemOld:DWORD
	LOCAL hTVItemNew:DWORD
	LOCAL 	h_tree:HTREEITEM
	LOCAL 	t_insert:TV_INSERTSTRUCT
	
	mov		eax, uMsg

	.if(eax==WM_COMMAND)
		;

	.elseif(uMsg==WM_MOUSEMOVE)
	
		; Handle a mouse-up event in the Treeview item
		; (not on a Treeview client area).
		.if(DragTargetLive == TRUE && DragMode == FALSE)
			; Custom mouse-up message
			invoke SendMessage, hPrjTrv, WM_LBUTTONUP_X, wParam, lParam
		.endif

		; We are draging
		.if(DragMode==TRUE)
			mov eax,lParam			; lParam contains the x/y coordinates of the cursor.
			and eax,0ffffh			; The high word (and eax, 1111,1111,1111,1111b)
			mov ecx,lParam			; The low word 
			shr ecx,16				; Shift (divide) right 16 positions. Add 16 leading "0"s
			mov tvhit.pt.x,eax	; The horizontal position of the drag image
			mov tvhit.pt.y,ecx	; The vertical position of the drag image		
			invoke ImageList_DragMove,eax,ecx		; Move the drag mage
			invoke ImageList_DragShowNolock,FALSE	; Hide the drag image											
			; Check if an item is hit as we drag over the Treeview nodes:													
			; Returns the handle to the tree-view item that occupies the specified 
			; point or NULL if no item occupies the point.
			invoke SendMessage,hPrjTrv,TVM_HITTEST,NULL,addr tvhit
			.if(eax!=NULL)
				; We are draging over a node, so highlite the node .
				invoke SendMessage,hPrjTrv,TVM_SELECTITEM,TVGN_DROPHILITE,eax
			.endif
			invoke ImageList_DragShowNolock,TRUE	; Show the drag image
		.endif


	; Mouse clicks in the Treeview window
	.elseif(eax==WM_LBUTTONDOWN)

		; Select the Treeview item that is currently being clicked.
		mov		eax, lParam
		and		eax, 0FFFFh		; 1111,1111,1111,1111b
		mov		tvhit.pt.x, eax
		mov		eax, lParam
		shr		eax, 16			; Shift (divide) right 16 positions. Add 16 leading "0"s
		mov		tvhit.pt.y, eax
		invoke SendMessage, hWin, TVM_HITTEST, 0, addr tvhit
		invoke SendMessage,hPrjTrv,TVM_SELECTITEM,TVGN_CARET,eax


		; Save the current image index
		; ...
		
		; Replace the target (drag-n-drop candidate) image 
		mov	tvi.imask, TVIF_IMAGE		; The Treeview image is changing
		m2m	tvi.hItem, hTVItem01			; The "target" node handle
		mov	tvi.iImage, 9					; The new "target" image index. zero based
		invoke SendMessage, hPrjTrv, TVM_SETITEM, 0, addr tvi
		mov DragTargetLive, TRUE			; The "target" node is now "live"


	; Custom WM_LBUTTONUP message
	.elseif(uMsg==WM_LBUTTONUP_X)
		; Note: Mouse-up in a Treeview item does not work as expected !!!
		; Revert the original "target" (drag-n-drop candidate) node images if we
		; have a WM_LBUTTONUP or WM_MOUSEMOVE
		.if(DragTargetLive == TRUE)
			mov	tvi.imask, TVIF_IMAGE	; The Treeview image is changing
			m2m	tvi.hItem, hTVItem01		; The "target" node handle
			mov	tvi.iImage, 1				; The old "target" image index. zero based
			invoke SendMessage, hPrjTrv, TVM_SETITEM, 0, addr tvi			
			mov DragTargetLive, FALSE		; The "target" node is no longer "live"
		.endif


	.elseif(uMsg==WM_LBUTTONUP)

		; Handle a mouse-up event in the Treeview client area 
		; (not on a Treeview item).
		.if(DragTargetLive == TRUE && DragMode == FALSE)
			; Custom mouse-up message if any drag targets are "live" 
			; (and user is not draging).
			invoke SendMessage, hPrjTrv, WM_LBUTTONUP_X, wParam, lParam
		.endif


		; Finish the drag operation
		.if(DragMode == TRUE)

			; Get the source Treeview item label
			invoke SendMessage, hPrjTrv, TVM_GETNEXTITEM, TVGN_CARET, NULL
			mov	hTVItemOld, eax
			mov	tvi.hItem, eax
			mov	tvi.imask, TVIF_PARAM or TVIF_TEXT
			lea	eax, buffItemSrc
			mov	tvi.pszText, eax
			mov	tvi.cchTextMax, sizeof buffItemSrc
			invoke SendMessage, hPrjTrv, TVM_GETITEM, 0, addr tvi
			;invoke	MessageBox, NULL, addr buffItemSrc, NULL, MB_OK

			; The drag item has been dropped
			invoke ImageList_DragLeave, hPrjTrv			; Unlocks the specified window ...
			invoke ImageList_EndDrag						; End the drag operation
			invoke ImageList_Destroy,hDragImageList	; Destroy the image list
			invoke SendMessage,hPrjTrv,TVM_GETNEXTITEM,TVGN_DROPHILITE,0 ; Get the drop target
			mov	 hTVItemNew, eax							; And save the new item handle
			invoke ReleaseCapture
			mov DragMode, FALSE

			; Get the target Treeview item label
			m2m	tvi.hItem, hTVItemNew
			mov	tvi.imask, TVIF_PARAM or TVIF_TEXT
			lea	eax, buffItemTgt
			mov	tvi.pszText, eax
			mov	tvi.cchTextMax, sizeof buffItemTgt
			invoke SendMessage, hPrjTrv, TVM_GETITEM, 0, addr tvi
			;invoke	MessageBox, NULL, addr buffItemTgt, NULL, MB_OK
			;1,get all properties of the source item
			invoke IfIncest,hPrjTrv,hTVItemNew,hTVItemOld
			.if eax
				invoke MessageBox, NULL, CStr("Incest!!!"), NULL, MB_OK
				ret
			.endif
			invoke InsertSon,hPrjTrv,hTVItemNew,hTVItemOld

		.endif


	.elseif(eax==WM_LBUTTONDBLCLK)
		;

	.elseif(eax==WM_RBUTTONDOWN)
		;
		;invoke MessageBox,NULL,CStr("left button down"),0,MB_OK
		invoke SendMessage, hPrjTrv, TVM_GETNEXTITEM, TVGN_CARET, NULL
		.if !eax
			invoke MessageBox,NULL,CStr("TVM_GETNEXTITEM error"),0,MB_OK
			ret
		.endif 
		
		
		mov h_tree,eax
		invoke SendMessage, hPrjTrv, TVM_GETNEXTITEM, TVGN_CHILD, h_tree
		.if !eax
			invoke MessageBox,NULL,CStr("TVM_GETNEXTITEM child error"),0,MB_OK
			ret
		.endif 
		mov h_tree,eax
		
		m2m	tvi.hItem, h_tree
			mov	tvi.imask, TVIF_PARAM or TVIF_TEXT
			lea	eax, buffItemTgt
			mov	tvi.pszText, eax
			mov	tvi.cchTextMax, sizeof buffItemTgt
			invoke SendMessage, hPrjTrv, TVM_GETITEM, 0, addr tvi
		invoke MessageBox,NULL ,addr buffItemTgt,0, MB_OK
		
		
	.elseif(uMsg==WM_NOTIFY)
		mov edi,lParam
		assume edi:ptr NM_TREEVIEW
		.if([edi].hdr.code==TVN_BEGINDRAG)						; We are draging (left mouse)
			invoke SendMessage,hPrjTrv,TVM_CREATEDRAGIMAGE,\; Create the (default) drag image
											0,[edi].itemNew.hItem
			mov hDragImageList,eax									; And save the drag image handle
			invoke ImageList_BeginDrag,hDragImageList,0,0,0	; Creates a temporary image list
			invoke ImageList_DragEnter,hPrjTrv,\				; Displays the drag image
												[edi].ptDrag.x,[edi].ptDrag.y
			invoke SetCapture,hWin									; Capture mouse to the Treeview
			mov DragMode,TRUE											; We are now draging
		.endif
		assume edi:nothing
	.endif

	invoke CallWindowProc, OldTreeViewProc, hWin, uMsg, wParam, lParam

	ret

TreeViewProc endp 



; Add TreeView Node
; Desc   : Inserts a new node in the Treeview control
;
; @param	hWin			The Treeview window handle (hPrjTrv)
; @param	lhPar			The node handle (hTVItem)
; @param	lhInsAfter	The item insert position (tvins.hInsertAfter)
; @param	pszText		The item text string (szNewProjectConfig = 'NewProjectConfig')
; @param	pidImg		Item icon id (item.iImage - non-selected state)
; @param	pidImgSel	Item icon id (iSelectedImage - selected state)
; @param	lParam		A 32-bit value to associate with the item
; @return	Returns the handle to the new item if successful or NULL otherwise
;
AddTreeViewNode proc hWin:HWND, lhPar:DWORD, lhInsAfter:DWORD, pszText:DWORD, \
							pidImgSel:DWORD, pidImg:DWORD, lParam:LPARAM
	LOCAL   tvins:TV_INSERTSTRUCT

	;invoke	MessageBox, NULL, NULL, NULL, MB_OK
	; Note: m2m is a memory-to-memory macro
	m2m		tvins.hParent     , lhPar			; Handle to the parent item
	m2m		tvins.hInsertAfter, lhInsAfter	; TVI_FIRST, VI_LAST, TVI_SORT
	m2m		tvins.item.lParam , lParam			; The node handle (hTVItem)
	mov		tvins.item._mask  , TVIF_TEXT or TVIF_IMAGE or TVIF_SELECTEDIMAGE or TVIF_PARAM
	m2m		tvins.item.pszText, pszText		; The item text string
	m2m		tvins.item.iImage , pidImg			; The normal (unselected) item image
	m2m		tvins.item.iSelectedImage, pidImgSel; The selected item image
	invoke SendMessage, hWin, TVM_INSERTITEM, 0, addr tvins

	; Return (eax) the handle to the new item if successful or NULL otherwise
	ret

AddTreeViewNode endp

InsertSon	proc htreeview,dest,src
	
	;1,get all properties of the source item
	LOCAL tvi:TV_ITEM
	LOCAL t_insert:TV_INSERTSTRUCT
	LOCAL  buffItemSrc[256]:byte
	LOCAL new_item:dword,son,sibling
	mov eax,dest
	.if eax==src
		ret
	.endif
	

	m2m	tvi.hItem,src
	mov	tvi.imask, TVIF_DI_SETITEM     or TVIF_IMAGE    or TVIF_SELECTEDIMAGE   or TVIF_STATE   or TVIF_TEXT 
	lea	eax, buffItemSrc
	mov	tvi.pszText, eax
	mov	tvi.cchTextMax, sizeof buffItemSrc
	invoke SendMessage, htreeview, TVM_GETITEM, 0, addr tvi
	
	;2.insert to the destination
	m2m t_insert.hParent,dest
	m2m t_insert.hInsertAfter,TVI_LAST
	
	invoke MemCopy,addr tvi,addr t_insert.item,sizeof tvi
	
	
	invoke SendMessage, htreeview, TVM_INSERTITEM, 0, addr t_insert
	
	mov new_item,eax
	
	invoke SendMessage,htreeview,TVM_GETNEXTITEM,TVGN_CHILD,src
	.if eax
		mov son,eax
		invoke InsertSon,htreeview,new_item,son	
		invoke SendMessage,htreeview,TVM_GETNEXTITEM,TVGN_NEXT,son
		.if eax
			mov sibling,eax
			invoke InsertSon,htreeview,new_item,sibling	
		.endif
		invoke SendMessage,htreeview,TVM_DELETEITEM,0,src
	.endif
	
	
	
	
	
	ret
	
InsertSon endp

IfIncest proc htreeview,son,father
	
	
	;to determin if the destination node is the son of the src
	;return 0 if no, return 1 if yes.
	LOCAL t_son,t_subling
	mov eax,son
	.if eax==father
		mov eax,1
		ret
	.endif
	
	invoke SendMessage,htreeview,TVM_GETNEXTITEM,TVGN_CHILD,father
	.if eax
		mov t_son,eax
		.if eax==son
			mov eax,1
			ret		
		.endif
		invoke IfIncest, htreeview,t_son,father
		.if eax==1
			ret
		.endif
		invoke SendMessage,htreeview,TVM_GETNEXTITEM,TVGN_NEXT,t_son
		.if eax
			mov t_subling,eax
			.if eax==son
				mov eax,1
				ret
			.endif	
		.endif
	.endif
	
	
	
	ret

IfIncest endp

end start
